// This is a CUDA program that does the following:
//
// 1. On the host, fill the indices array with the index of each rectangle
// 2. Copy the indices array from the host to the device
// 3. Use the device to run a kernel that:
//    a. Uses the indices array to determine the left x-coordinate of
//       each rectangle (float const x = deviceIndices[threadId] * width)
//    b. Uses the left x-coordinate to determine the height of the rectangle
//       (float const height = sqrtf(1.0 - x * x)
//    b. Uses the width and height to determine the area
//       (deviceAreas[threadId] = width * height)
// 4. Copy the areas array from the device to the host
// 5. On the host, sum the areas and multiply by 4 to calculate an approximation
//    of pi
// 6. On the host, print the resulting approximation of pi
//
// Author: Aaron Weeden, Shodor, 2016

// Import library so we can call printf()
#include <stdio.h>

// Import library so we can call exit(), malloc(), free(), etc.
#include <stdlib.h>

// Define the number of rectangles
#define RECT_COUNT 10

// Define the number of bytes in each array (depending on the type)
#define INT_BYTE_COUNT   ((RECT_COUNT) * sizeof(int))
#define FLOAT_BYTE_COUNT ((RECT_COUNT) * sizeof(float))

// Define the number of CUDA threads in each CUDA warp (group of threads that
// execute instructions in lock-step)
#define THREADS_PER_WARP 32

// Define the maximum number of CUDA warps in each CUDA block
#define MAX_WARPS_PER_BLOCK 16

// Define the number of CUDA threads in each CUDA block
#define THREADS_PER_BLOCK ((THREADS_PER_WARP) * (MAX_WARPS_PER_BLOCK))

// Define the number of CUDA blocks in each CUDA grid
#define BLOCKS_PER_GRID 1

// Declare functions that will be defined later
void TryMalloc(void * const err);
void TryCuda(cudaError_t const err);
__global__ void CalcAreas(int const * const indices, float * const areas,
                          float const width, int const count);

// Start the program
int main()
{
  // Define a variable for the width of a rectangle
  float width = (float)1.0 / RECT_COUNT;

  // Declare variables for the host and device arrays
  int   * hostIndices;
  float * hostAreas;
  int   * deviceIndices;
  float * deviceAreas;

  // Allocate memory for the host arrays
  TryMalloc(hostIndices = (int*)  malloc(INT_BYTE_COUNT));
  TryMalloc(hostAreas   = (float*)malloc(FLOAT_BYTE_COUNT));

  // Allocate memory for the device arrays
  TryCuda(cudaMalloc((void**)&deviceIndices, INT_BYTE_COUNT));
  TryCuda(cudaMalloc((void**)&deviceAreas,   FLOAT_BYTE_COUNT));

  // On the host, fill the indices array with the index of each rectangle
  for (int i = 0; i < RECT_COUNT; i++)
  {
    hostIndices[i] = i;
  }

  // Copy the indices array from the host to the device
  TryCuda(cudaMemcpy(deviceIndices, hostIndices, INT_BYTE_COUNT,
                     cudaMemcpyHostToDevice));

  // Use the device to run a kernel
  CalcAreas<<<BLOCKS_PER_GRID, THREADS_PER_BLOCK>>>(deviceIndices, deviceAreas,
                                                    width, RECT_COUNT);

  // Copy the areas array from the device to the host
  TryCuda(cudaMemcpy(hostAreas, deviceAreas, FLOAT_BYTE_COUNT,
                     cudaMemcpyDeviceToHost));

  // On the host, sum the areas and multiply by 4 to calculate an approximation
  // of pi
  float pi = 0.0;
  for (int i = 0; i < RECT_COUNT; i++)
  {
    pi += hostAreas[i];
  }
  pi *= 4.0;

  // On the host, print the resulting approximation of pi
  printf("Result:\tpi = %f\n", pi);

  // De-allocate memory for the device arrays
  TryCuda(cudaFree(deviceAreas));
  TryCuda(cudaFree(deviceIndices));

  // De-allocate memory for the host arrays
  free(hostAreas);
  free(hostIndices);

  return 0;
}

// Define a function to check whether a malloc() call was successful
void TryMalloc(void * const err)
{
  if (err == NULL)
  {
    fprintf(stderr, "malloc error\n");
    exit(EXIT_FAILURE);
  }
}

// Define a function to check whether a CUDA call was successful
void TryCuda(cudaError_t const err)
{
  if (err != cudaSuccess)
  {
    fprintf(stderr, "CUDA Error: %s\n", cudaGetErrorString(err));
    exit(EXIT_FAILURE);
  }
}

// Define a function which will be executed on a CUDA device
__global__ void CalcAreas(int const * const deviceIndices,
                          float * const deviceAreas, float const width,
                          int const count)
{
  // Calculate the unique ID for the current CUDA thread
  int const threadId = blockIdx.x * blockDim.x + threadIdx.x;

  // All threads whose thread ID is >= count will NOT do the following, thus
  // avoiding writing into un-allocated space.
  if (threadId < count)
  {
    float const x = deviceIndices[threadId] * width;
    float const height = sqrtf(1.0 - x * x);
    deviceAreas[threadId] = width * height;
  }
}

